use crate::{
    file_watcher::FileWatcher,
    prelude::*,
    state::{NewStateError, State as AppState},
    static_files,
};
use axum::{
    extract::State,
    http::StatusCode,
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use minijinja::context;
use tower_livereload::LiveReloadLayer;

use thiserror::Error;

#[derive(Error, Debug)]
pub enum NewRouterError {
    #[error("new state error: {0}")]
    State(#[from] NewStateError),

    #[error("watcher error: {0}")]
    Watcher(#[from] notify::Error),
}

#[derive(Error, Debug)]
pub enum ReqError {
    #[error("template error: {0}")]
    Template(#[from] minijinja::Error),
}

impl IntoResponse for ReqError {
    fn into_response(self) -> axum::http::Response<axum::body::Body> {
        error!("webserver error: {:?}", self);
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            // TODO: don't expose raw errors over the internet?
            format!("internal server error: {}", self),
        )
            .into_response()
    }
}

pub type ReqResult<T> = Result<T, ReqError>;

pub async fn router(
    with_watchers: bool,
) -> Result<(Router, Vec<Option<FileWatcher>>), NewRouterError> {
    let state = AppState::try_new().await?;

    let live_reload_layer: Option<LiveReloadLayer> = if with_watchers {
        Some(LiveReloadLayer::new())
    } else {
        None
    };

    let orl = || {
        if let Some(lr) = &live_reload_layer {
            Some(lr.reloader())
        } else {
            None
        }
    };

    let template_file_watcher = state.clone().templates.start_watcher(orl()).await?;
    let (static_file_service, static_file_watcher) = static_files::router(orl())?;

    let mut result = Router::new()
        .route("/", get(index))
        .route("/about", get(about))
        .nest_service("/static", static_file_service)
        .with_state(state.clone());

    if let Some(lr) = live_reload_layer {
        result = result.clone().layer(lr);
    }

    let watchers = vec![template_file_watcher, static_file_watcher];

    Ok((result, watchers))
}

async fn index(State(state): State<AppState>) -> ReqResult<Html<String>> {
    Ok(Html(
        state
            .templates
            .render("pages/index.jinja.html", context!())
            .await?,
    ))
}
async fn about(State(state): State<AppState>) -> ReqResult<Html<String>> {
    Ok(Html(
        state
            .templates
            .render("pages/about.jinja.html", context!())
            .await?,
    ))
}
